/* 
 *  rocks/hb.c
 *
 *  The heartbeat connection failure detection mechanism.
 *
 *  Copyright (C) 2001 Victor Zandy
 *  See COPYING for distribution terms.
 */

/*
  We create a auxiliary TCP connection for the heartbeat.  Each end of
  the current data connection creates a listening socket bound to an
  arbitrary port.  It exchanges the port number to its peer, and then
  accepts incoming connections on the socket and simulataneously
  attempts to connect to the peer.  1 of the 2 possible connections is
  selected for the heartbeat connection.

  There are numerous problems with sharing heartbeat over the data connection:
      - Some systems, such as Linux, copy old out-of-band urgent data
        in the receive queue into the data stream if it is not consumed
        soon enough.
      - Neither urgent data nor urgent notification are sent when the
        send queue is full on Linux.
      - Race-free implementations are complicated.  */

#include <sys/time.h>
#include <sys/types.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>

#include "rs.h"
#include "rstypes.h"
#include "log.h"

struct hb {
	int period;		/* Beat period (multiple of alarm period) */
	int count;		/* Beats that have occurred within period */
	int limit;		/* Missed beats limit */
	int missed;		/* Current number of missed beats */
	int s;   		/* Connection to peer hb socket (plain) */
	rs_t rs;		/* Pointer to owning rock */
};

/* These are NULL unless there is a HB for the descriptor. */
static hb_t hbs[RS_MAXFD];

hb_t
rs_new_heartbeat(rs_t rs)
{
	hb_t hb;
	hb = (hb_t) malloc(sizeof(struct hb));
	if (!hb)
		return NULL;
	hb->limit = rs_opt_max_alarm_misses;
	hb->period = 1;
	hb->count = 0;
	hb->missed = 0;
	hb->s = -1;
	hb->rs = rs;
	return hb;
}

void
rs_free_heartbeat(hb_t hb)
{
	rs_mode_native();
	close(hb->s);
	rs_mode_pop();
	free(hb);
}

static hb_t
hb_lookup(int s)
{
	if (s >= RS_MAXFD || s < 0)
		return NULL;
	return hbs[s];
}

int
rs_1of2_client(int *passivep, int *activep, fd_set *fds)
{
	int pick = -1;
	char byte;
	int passive = *passivep;
	int active = *activep;

	if (active >= 0 && FD_ISSET(active, fds)) {
		if (0 >= rs_xread(active, &byte, sizeof(byte), 0)) {
			close(active);
			*activep = -1;
		} else {
			pick = active;
			close(passive);
		}
	} 
	if (pick < 0 && passive >= 0 && FD_ISSET(passive, fds)) {
		if (0 >= rs_xread(passive, &byte, sizeof(byte), 0)) {
			close(passive);
			*passivep = -1;
		} else {
			pick = passive;
			close(active);
		}
	}
	return pick;
}

int
rs_1of2_server(int *passivep, int *activep, fd_set *fds)
{
	int pick = -1;
	char byte;
	int passive = *passivep;
	int active = *activep;

	if (active >= 0) {
		if (0 > rs_xwrite(active, &byte, sizeof(byte))) {
			close(active);
			*activep = -1;
		} else {
			pick = active;
			close(passive);
		}
	} 
	if (pick < 0 && passive >= 0) {
		if (0 > rs_xwrite(passive, &byte, sizeof(byte))) {
			close(passive);
			*passivep = -1;
		} else {
			pick = passive;
			close(active);
		}
	}
	return pick;
}

int
rs_1of2(int *passive, int *active, fd_set *fds, rs_role role)
{
	if (RS_ROLE_SERVER == role)
		return rs_1of2_server(passive, active, fds);
	else						  
		return rs_1of2_client(passive, active, fds);
}

/* ROCK is established data connection descriptor.
   HB is allocated, initialized HB structure.
   Establish HB connection and update HB accordingly.
   ROLE is used to decide ties. */
int
rs_hb_establish(int rock, hb_t hb, rs_role role)
{
	int rv, len;
	int ls = -1, cs = -1, ns = -1; /* Init for safe closes at "out" */
	int pick;             /* Selected connection descriptor */ 
	struct sockaddr_in laddr, paddr;
	int connected;

	rs_mode_native();

	/* Bind a new listening socket with IP address of ROCK */
	len = sizeof(laddr);
	if (0 > getsockname(rock, (struct sockaddr *)&laddr, &len))
		goto out;
	ls = socket(AF_INET, SOCK_STREAM, 0);
	if (0 > ls)
		goto out;
	len = sizeof(laddr);
	laddr.sin_port = htons(0);
	rv = bind(ls, (struct sockaddr *) &laddr, sizeof(laddr));
	if (0 > rv)
		goto out;
	rv = listen(ls, 1);
	if (0 > rv)
		goto out;

	/* Non blocking avoids timing problem described in
	   Stevens, Network Programming 1, 15.6 */
	if (0 > rs_nonblock(ls, 1))
		goto out;

	/* Exchange socket addresses with peer - network order */
	len = sizeof(laddr);
	if (0 > getsockname(ls, (struct sockaddr *)&laddr, &len))
		goto out;
	if (0 >= rs_xwrite(rock, &laddr, sizeof(laddr)))
		goto out;
	if (0 >= rs_xread(rock, &paddr, sizeof(paddr), 0))
		goto out;

	/* Connect */
	cs = socket(AF_INET, SOCK_STREAM, 0);
	if (0 > cs)
		goto out;
	if (0 > rs_nonblock(cs, 1))
		goto out;
	rv = connect(cs, (struct sockaddr *)&paddr, sizeof(paddr));
	if (0 > rv && EINPROGRESS != errno)
		goto out;

	connected = 0;
	pick = -1;
	while (pick < 0) {
		fd_set fds;
		struct timeval tv;
		int max;

		FD_ZERO(&fds);
		max = 0;
		if (!connected && cs > 0) {
			FD_SET(cs, &fds);
			max = cs;
		}
		if (ns < 0 && ls > 0) {
			FD_SET(ls, &fds);
			max = MAX(max, ls);
		}
		if (ns >= 0 && RS_ROLE_CLIENT == role) {
			FD_SET(ns, &fds);
			max = MAX(max, ns);
		}

		tv.tv_sec = 0;
		tv.tv_usec = 5000;
		rv = select(1+max, &fds, NULL, NULL, &tv);
		if (0 > rv && EINTR == errno)
			continue;   /* interrupt */
		if (0 == rv)
			continue;   /* timeout */
		if (0 > rv)
			goto out;   /* error */

		len = sizeof(paddr);
		if (FD_ISSET(ls, &fds)) {
			ns = accept(ls, (struct sockaddr *)&paddr, &len);
			if (ns >= 0 && RS_ROLE_SERVER == role)
				pick = rs_1of2(&ns, &cs, &fds, role);
			continue;
		}
		if (!connected && FD_ISSET(cs, &fds)) {
			len = sizeof(paddr);
			if (0 > getpeername(cs, (struct sockaddr *)&paddr,
					    &len)) {
				FD_CLR(cs, &fds);
				cs = -1; /* connect failed */
			} else {
				connected = 1;
				if (RS_ROLE_SERVER == role)
					pick = rs_1of2(&ns, &cs, &fds, role);
			}
			continue;
		}
		/* One of CS or NS is ready for reading */
		if (RS_ROLE_CLIENT == role)
			pick = rs_1of2(&ns, &cs, &fds, role);
	}
	close(ls);
	if (0 > fcntl(pick, F_SETOWN, getpid()))
		assert(0);
	hb->s = pick;
	hbs[hb->s] = hb;
	rs_mode_pop();
	rs_log("Heartbeat connection established for <%p>.", hb->rs);
	return 0;
out:
	close(ls);
	close(cs);
	rs_mode_pop();
	rs_log("Cannot establish heartbeat connection for <%p>.", hb->rs);
	return -1;
}

int
rs_hb_cancel(hb_t hb)
{
	if (hb->s >= 0) {
		hbs[hb->s] = NULL;
		close(hb->s);
	}
	hb->s = -1;
	hb->count = 0;
	hb->missed = 0;
	return 0;
}

/* Return 0 if sent probe okay.  Return -1 if the connection went away. */
static int
send_probe(int sd)
{
	int rv, e;
send_probe_retry:
	rv = send(sd, "x", 1, MSG_OOB);
	if (rv == 1)
		return 0;
	assert(rv != 0);
	e = errno;
	switch(e) {
	case EINTR:
		goto send_probe_retry;
		break;
	case EAGAIN:
	case ENOSPC:
	case EIO:
	case EFAULT:
		/* These shouldn't happen */
		assert(0);
		break;
	case EBADF:
	case EINVAL:  /* Linux can do this if you remove the nic */
	default:
 	        /* The connection went away */
		break;
	}
	return -1;
}

static void
handle_sigalrm(int sig)
{
	fd_set fdset;
	rs_t rs;
	int max;
	int sd;

	rs_mode_native();
	FD_ZERO(&fdset);
	max = rs_fdset(&fdset);
	for (sd = 0; sd < max; sd++) {
		if (!FD_ISSET(sd, &fdset))
			continue;
		rs = rs_lookup(sd);
		assert(rs);
		/* Count beats toward this heartbeat's period */
		rs->hb->count++;
		if (rs->hb->count < rs->hb->period)
			continue;
		
		/* Period expired */
		rs->hb->count = 0;
		rs->hb->missed++;
		
		/* Send probe */
		if (0 > send_probe(rs->hb->s)) {
			rs_begin_reconnect(rs);
			continue;
		}
		
		/* Check missed probe */
		if (rs->hb->missed >= rs->hb->limit)
			rs_begin_reconnect(rs);
		else if (rs->hb->missed > 1)
			rs_log("missed heartbeat <%p> (suspending after %d more)",
			       rs, rs->hb->limit - rs->hb->missed);
	}
	rs_mode_pop();
}

static void
handle_sigurg(int sig)
{
	fd_set fdsetr, fdseth, fds;
	int max, maxh, n, sd, rv;
	struct timeval tv;
	char x;
	
	rs_mode_native();

retry_all:
	FD_ZERO(&fdsetr);
	FD_ZERO(&fds);
	max = rs_fdset(&fdsetr);

	/* FIXME: Rewrite as something similar to rs_fdset */
	maxh = 0;
	for (sd = 0; sd < max; sd++)
		if (FD_ISSET(sd, &fdsetr)) {
			rs_t rs = rs_lookup(sd);
			assert(rs);
			FD_SET(rs->hb->s, &fds);
			if (rs->hb->s > maxh)
				maxh = rs->hb->s;
		}
	maxh++;
retry_select:
	memcpy(&fdseth, &fds, sizeof(fdseth));
	tv.tv_sec = 0;
	tv.tv_usec = 0;
	n = select(maxh, NULL, NULL, &fdseth, &tv);
	if (0 > n) {
		if (errno == EBADF
		    && 0 > rs_recover_bad_rocks(maxh, &fds))
			assert(0);
		goto retry_select;
	}

	for (sd = 0; n > 0 && sd < maxh; sd++)
		if (FD_ISSET(sd, &fdseth)) {
			hb_t hb;
			hb = hb_lookup(sd);
			assert(hb);
			rv = recv(hb->s, &x, 1, MSG_OOB);
			if (0 > rv) {
				rs_begin_reconnect(hb->rs);
				/* Descriptors in fdseth may be invalid
				   after reconnection. */
				goto retry_all;
			} else
				hb->missed = 0;
			n--;
		}
	rs_mode_pop();
}

void 
rs_stop_heartbeat(sigset_t *oset)
{
	sigset_t set;
	sigemptyset(&set);
	sigaddset(&set, SIGALRM);
	sigprocmask(SIG_BLOCK, &set, oset);
}

void
rs_resume_heartbeat(sigset_t *oset)
{
	sigprocmask(SIG_SETMASK, oset, NULL);
}

int
rs_init_heartbeat()
{
	struct sigaction sa;
	struct itimerval it;
	int ret = 0;

	sigfillset(&sa.sa_mask);
	sigdelset(&sa.sa_mask, SIGTERM);
	sigdelset(&sa.sa_mask, SIGINT);
	sa.sa_flags = SA_RESTART;
	sa.sa_restorer = 0;
	sa.sa_handler = handle_sigurg;
	if (0 > sigaction(SIGURG, &sa, NULL))
		return -1;
	sa.sa_handler = handle_sigalrm;
	if (0 > sigaction(SIGALRM, &sa, NULL))
		return -1;

	it.it_value.tv_sec = it.it_interval.tv_sec = rs_opt_alarm_period;
	it.it_value.tv_usec = it.it_interval.tv_usec = 0;
	rs_mode_native();
	ret = setitimer(ITIMER_REAL, &it, NULL);
	rs_mode_pop();
	return ret;
}

int
rs_setitimer(int which, const struct itimerval *value, struct itimerval *ov)
{
	return 0;
}
